/* audio.c
 * - audio processing
 *
 * Copyright (c) 2003 Juho Vh-Herttua <juhovh@iki.fi>
 * Based on audio.c of ices2 by Michael Smith <msmith@labyrinth.net.au>
 *
 * This program is distributed under the terms of the GNU General
 * Public License, version 2. You may use, modify, and redistribute
 * it under the terms of this license. A copy should be included
 * with this source.
 */

#ifdef HAVE_CONFIG_H
 #include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>

#include "audio.h"
#include "resample.h"

#define MAX_CHANNELS 2

/* Allocate buffer pointers */
process_state *process_initialise(void)
{
    process_state *state = g_malloc0(sizeof(process_state));
    int failed = 1;

    do 
    {
        if (state==NULL)
            break;
        if ((state->bitbuf = g_malloc0(sizeof(gint8 *))) == NULL)
            break;
        if ((state->convbuf = g_malloc0(MAX_CHANNELS*sizeof(gfloat *))) == NULL)
            break;
        if ((state->chanbuf = g_malloc0(MAX_CHANNELS*sizeof(gfloat *))) == NULL)
            break;        
        if ((state->buffers = g_malloc0(MAX_CHANNELS*sizeof(gfloat *))) == NULL)
            break;
        failed = 0;
    }
    while (0); /* not a loop */
        
    if (failed)
    {
        g_print("Couldn't initialise audio processor due to memory allocation failure\n");
        process_clear (state);
        return NULL;
    }
    
    state->resample = 0;

    return state;
}

/* Update processor settings */
void process_update(process_state *state, gint lowbits, gint usigned, gint be, gint inchannels, gint outchannels, gint infreq, gint outfreq)
{
    state->lowbits = lowbits;
    state->usigned = usigned;
    state->bigendian = be;
    state->inchannels = inchannels;
    state->outchannels = outchannels;

    if (infreq != outfreq) {
        if (resampler_init(&state->resampler, outchannels, outfreq, infreq, RES_END)) {
            g_print("Couldn't initialise resampler to specified frequency\n");
            return;
        }
        else {
            state->resample = 1;
        }
    }
    else state->resample = 0;
}

/* Finish processing, flush resampler */
gint process_finish(process_state *s, gfloat ***dest)
{
    gint ret;
    
    if (!s->resample)
        return 0;

    ret = resampler_drain(&s->resampler, s->buffers);

    if(ret > s->bufsize) {
        g_print("Fatal error in resampler: buffers too small");
        return 0;
    }

    s->buffill = ret;
    
    resampler_clear(&s->resampler);
    
    *dest = s->buffers;
    return s->buffill;
}

/* Clear processor struct */
void process_clear(process_state *s)
{
    gint c;

    if(s) {
        if(s->bitbuf) {
            g_free(s->bitbuf);
        }
        if(s->convbuf) {
            for(c=0; c<s->inchannels; c++)
                if (s->buffers[c])
                    g_free(s->convbuf[c]);
            g_free(s->convbuf);
        }
        if(s->chanbuf) {
            for(c=0; c<s->outchannels; c++)
                if (s->chanbuf[c])
                    g_free(s->chanbuf[c]);
            g_free(s->chanbuf);
        }
        if(s->buffers) {
            for(c=0; c<s->outchannels; c++)
                if (s->buffers[c])
                    g_free(s->buffers[c]);
            g_free(s->buffers);
        }
        g_free(s);
    }
}

/* Process audio */
gint process_audio(process_state *s, gint8 *buf, gint buflen, gfloat ***dest)
{
    if (s->lowbits) {
        if (s->usigned)
             convert_usigned_lowbits(s, buf, buflen);
        convert_lowbits(s, buf, buflen);
        buf = (gint8 *) s->bitbuf;
        buflen = s->bitbuflen;
    }
    else if (s->usigned)
        convert_usigned(s, buf, buflen);
    if (s->bigendian)
        convert_endianness(s, buf, buflen);
    
    /* Now the buffer buf should be 16bit signed low-endian int in all cases */
    convert_float(s, buf, buflen);
    if (s->inchannels != s->outchannels) {
        convert_channels(s, s->convbuf, s->convbuflen);
        if (s->resample) {
            resample_buffer(s, s->chanbuf, s->chanbuflen);
            *dest = s->buffers;
            return s->buffill;
        }
        else {
            *dest = s->chanbuf;
            return s->chanbuflen;
        }
    }    
    else if (s->resample) {
        resample_buffer(s, s->convbuf, s->convbuflen);
        *dest = s->buffers;
        return s->buffill;
    }
    else {
        *dest = s->convbuf;
        return s->convbuflen;
    }
}

/* 8bit unsigned to signed conversion */
void convert_usigned_lowbits(process_state *s, gint8 *buf, gint buflen)
{
    gint i;
    guint8 *src = (guint8 *) buf;
    
    for (i=0; i<buflen; i++) {
        buf[i] = src[i]-128;
    }
}

/* 16bit unsigned to signed conversion */
void convert_usigned(process_state *s, gint8 *buf, gint buflen)
{
    gint i;
    guint16 *src = (guint16 *) buf;
    gint16 *dest = (gint16 *) buf;
    
    for (i=0; i<buflen; i++) {
        dest[i] = src[i]-32768;
    }
}

/* 8bit->16bit conversion */
void convert_lowbits(process_state *s, gint8 *buf, gint buflen)
{
    gint i;
    
    if(buflen*2 > s->bitbuflen) {
        void *tmp = g_realloc(s->bitbuf, buflen*2 * sizeof(gint8));
        if (tmp==NULL) {
            g_print("Error in converting from 8bit to 16bit\n");
            return;
        }
        s->bitbuf = tmp;
    }
    s->bitbuflen = buflen*2;
    
    for (i=0; i < buflen; i++) {
        s->bitbuf[i*2+1] = buf[i];
        s->bitbuf[i*2] = 0;
    }
}

/* Convert low endian to big endian and vice versa */
void convert_endianness(process_state *s, gint8 *buf, gint buflen) {
    gint i;
    gint8 temp;
    
    for (i=0; i<buflen/2; i++) {
        temp = buf[i*2];
        buf[i*2] = buf[i*2+1];
        buf[i*2+1] = temp;
    }
}

/* Convert integer buffer to floating point */
void convert_float(process_state *s, gint8 *buf, gint buflen)
{
    gint c,i;
    buflen /= 2*s->inchannels; /* bytes -> samples conversion */

    if(s->convbuflen < buflen) {
        for(c=0; c < s->inchannels; c++)
            s->convbuf[c] = g_realloc(s->convbuf[c], buflen * sizeof(gfloat));
    }

    for(i=0; i < buflen; i++) {
        for(c=0; c < s->inchannels; c++) {
            s->convbuf[c][i] = ((buf[2*(i*s->inchannels + c) + 1]<<8) |
                                (0x00ff&(gint)buf[2*(i*s->inchannels + c)]))/
                                 32768.f;
        }
    }
    s->convbuflen = buflen;
}

/* Convert mono to stereo or vice versa */
void convert_channels(process_state *s, gfloat **buf, gint buflen) {
    gint i, c;
    
    if(s->chanbuflen < buflen) {
        for(c=0; c < s->outchannels; c++)
            s->chanbuf[c] = g_realloc(s->chanbuf[c], buflen * sizeof(gfloat));
    }
    if (s->inchannels == 1 && s->outchannels == 2) {
        for (i=0; i<buflen; i++) {
            s->chanbuf[0][i] = buf[0][i];
            s->chanbuf[1][i] = buf[0][i];
        }
    }
    else if (s->inchannels == 2 && s->outchannels == 1) {
        for (i=0; i<buflen; i++) {
            s->chanbuf[0][i] = (buf[0][i] + buf[1][i]) * 0.5;
        }
    }
    s->chanbuflen = buflen;
}

/* Resample floating point buffer */
void resample_buffer(process_state *s, gfloat **buf, gint buflen)
{
    gint c;
    gint res;

    s->buffill = resampler_push_check(&s->resampler, buflen);
    if(s->buffill <= 0) {
        g_print("Fatal reencoding error: resampler_push_check returned %d, input: %i\n", 
                s->buffill, buflen);
    }

    if(s->bufsize < s->buffill) {
        s->bufsize = s->buffill;
        for(c=0; c<s->outchannels; c++)
            s->buffers[c] = g_realloc(s->buffers[c], s->bufsize * sizeof(gfloat));
    }

    if((res = resampler_push(&s->resampler, s->buffers, (gfloat const **)buf, buflen))
            != s->buffill) {
        g_print("Internal error in resampling: returned number of samples %d"
                   ", expected %d\n", res, s->buffill);
        s->buffill = res;
        return;
    }
}
