/* gui.c
 * - configuration window and callbacks
 *
 * Copyright (c) 2003 Juho Vh-Herttua <juhovh@iki.fi>
 *
 * This program is distributed under the terms of the GNU General
 * Public License, version 2. You may use, modify, and redistribute
 * it under the terms of this license. A copy should be included
 * with this source.
 */

#include <stdlib.h>
#include <stdio.h>
#include <shout/shout.h>
#include <gtk/gtk.h>
#include "oggcaster.h"
#include "gui.h"
#include "stream.h"

#define SAMPLERATES 6

static oc_widgets *widgets;
static gint messageid = 0;

static gint oc_samplerates[SAMPLERATES]={
	11025,
	22050,
	24000,
	32000,
	44100,
	48000
};

static gchar *messages[6]={
    "Disconnected from server",
    "Connection succesful",
    "Error connecting to server",
    "Error initialising stream",
    "Error initialising encoder",
    "Connection lost to server"
};

void destroy(GtkWidget *widget, gpointer data)
{
    g_free(widgets);
    widgets = NULL;
}

void set_managed_cb(GtkWidget *w, gpointer data)
{
    widgets->managed = GPOINTER_TO_INT(data);
}

void set_samplerate_cb(GtkWidget *w, gpointer data)
{
    widgets->samplerate = GPOINTER_TO_INT(data);
}

void set_channels_cb(GtkWidget *w, gpointer data)
{
    widgets->channels = GPOINTER_TO_INT(data);
}

void connect_button_cb(GtkWidget *w, gpointer data)
{
    gtk_widget_set_sensitive(GTK_WIDGET(widgets->connect_button), FALSE);
    pthread_mutex_lock(&widgets->state->runlock);
    if (!widgets->state->running) {
        gtk_label_set(GTK_LABEL (GTK_BIN(widgets->connect_button)->child), "Disconnect");
        get_values(widgets->state->stream);
        stream_initialise(widgets->state);
    }
    else {
        gtk_label_set(GTK_LABEL (GTK_BIN(widgets->connect_button)->child), "Connect");
        widgets->state->running = FALSE;
        stream_queue(widgets->state, NULL, 0);
    }
    pthread_mutex_unlock(&widgets->state->runlock);
}

void connect_button_activate(gint msgid)
{
    messageid = msgid;
    
    if (widgets) {
        if (GTK_WIDGET_IS_SENSITIVE(GTK_WIDGET(widgets->connect_button)) && msgid == 0) return;
        
        pthread_mutex_lock(&widgets->state->runlock);
        gdk_threads_enter();
        gtk_entry_set_text(GTK_ENTRY (widgets->status_entry), messages[messageid]);
        if (messageid == 1) {
            widgets->state->running = TRUE;
            widgets->state->previous_song = -1;
        }
        else
            gtk_label_set(GTK_LABEL (GTK_BIN(widgets->connect_button)->child), "Connect");
        gtk_widget_set_sensitive(GTK_WIDGET(widgets->connect_button), TRUE);
        gdk_threads_leave();
        pthread_mutex_unlock(&widgets->state->runlock);
    }     
}

void apply_button_cb(GtkWidget *w, gpointer data)
{
    pthread_mutex_lock(&widgets->state->runlock);
    if (!widgets->state->new_stream) {
        get_values(widgets->state->stream);
        widgets->state->new_stream = TRUE;
    }
    pthread_mutex_unlock(&widgets->state->runlock);
}

void mute_button_cb(GtkWidget *w, gpointer data)
{
    pthread_mutex_lock(&widgets->state->runlock);
    if (!widgets->state->muted)
        widgets->state->muted = 1;
    else widgets->state->muted = 0;
    pthread_mutex_unlock(&widgets->state->runlock);
}

void ok_button_cb(GtkWidget *w, gpointer data)
{
    get_values(widgets->state->saved_stream);
}

void configure(oc_state_info *state)
{
    GtkWidget *configure_window;
    GtkWidget *notebook;
    GtkWidget *main_vbox, *format_vbox, *server_vbox;
    GtkWidget *format_frame, *info_frame, *server_frame, *encoding_frame;
    GtkWidget *buttons_vbox;
    GtkWidget *notebook_label1, *notebook_label2;
    
    /* FIXME: We should raise the window on top here */
    if (widgets) {
        return;
    }
    
    widgets = g_malloc0(sizeof(oc_widgets));

    widgets->state = state;
    widgets->state->stream = state->stream;
    
    configure_window = gtk_window_new (GTK_WINDOW_DIALOG);
    
    gtk_object_set_data (GTK_OBJECT (configure_window), "configure_window", configure_window);
    gtk_window_set_title (GTK_WINDOW (configure_window), "Configure OggCaster Plugin");
    gtk_window_set_policy(GTK_WINDOW (configure_window), FALSE, FALSE, FALSE);
    
    gtk_signal_connect (GTK_OBJECT (configure_window), "delete_event",
                        GTK_SIGNAL_FUNC (gtk_widget_destroy), GTK_OBJECT (configure_window));
    gtk_signal_connect (GTK_OBJECT (configure_window), "destroy",
                        GTK_SIGNAL_FUNC (destroy), NULL);
    
    gtk_container_set_border_width (GTK_CONTAINER (configure_window), 10);
    
    main_vbox = gtk_vbox_new (FALSE, 0);
    gtk_container_add (GTK_CONTAINER (configure_window), main_vbox);
    gtk_widget_show (main_vbox);
    
    notebook = gtk_notebook_new ();
    gtk_box_pack_start (GTK_BOX (main_vbox), notebook, TRUE, TRUE, 0);
    gtk_widget_show (notebook);
    
    format_vbox = gtk_vbox_new (FALSE, 5);
    gtk_container_add (GTK_CONTAINER (notebook), format_vbox);
    gtk_widget_show (format_vbox);
    
    encoding_frame = create_encoding_frame();
    gtk_container_set_border_width (GTK_CONTAINER (encoding_frame), 5);
    gtk_box_pack_start (GTK_BOX (format_vbox), encoding_frame, FALSE, FALSE, 0);
    gtk_widget_show (encoding_frame);
    
    format_frame = create_format_frame();
    gtk_box_pack_start (GTK_BOX (format_vbox), format_frame, FALSE, FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (format_frame), 5);
    gtk_widget_show (format_frame);
    
    notebook_label1 = gtk_label_new ("Audio Format");
    gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook), gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), 0), notebook_label1);
    gtk_widget_show (notebook_label1);
    
    server_vbox = gtk_vbox_new (FALSE, 5);
    gtk_container_add (GTK_CONTAINER (notebook), server_vbox);
    gtk_widget_show (server_vbox);
    
    server_frame = create_server_frame();
    gtk_container_set_border_width (GTK_CONTAINER (server_frame), 5);
    gtk_box_pack_start (GTK_BOX (server_vbox), server_frame, FALSE, FALSE, 0);
    gtk_widget_show (server_frame);
    
    info_frame = create_info_frame();
    gtk_container_set_border_width (GTK_CONTAINER (info_frame), 5);
    gtk_box_pack_start (GTK_BOX (server_vbox), info_frame, FALSE, FALSE, 0);
    gtk_widget_show (info_frame);
    
    notebook_label2 = gtk_label_new ("Server Options");
    gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook), gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), 1), notebook_label2);
    gtk_widget_show (notebook_label2);
        
    widgets->status_entry = gtk_entry_new ();
    gtk_entry_set_editable(GTK_ENTRY(widgets->status_entry), FALSE);
    gtk_box_pack_start (GTK_BOX (main_vbox), widgets->status_entry, FALSE, FALSE, 10);
    gtk_widget_show (widgets->status_entry);
    gtk_entry_set_text(GTK_ENTRY (widgets->status_entry), messages[messageid]);
    
    buttons_vbox = create_buttons();
    gtk_signal_connect_object( GTK_OBJECT(widgets->ok_button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT (configure_window));
    gtk_signal_connect_object( GTK_OBJECT(widgets->cancel_button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT (configure_window));
    gtk_box_pack_start (GTK_BOX (main_vbox), buttons_vbox, FALSE, FALSE, 0);
    gtk_widget_show (buttons_vbox);
    
    set_values(widgets->state->saved_stream);
    
    gtk_widget_show (configure_window);
}

void set_values(oc_stream_info *stream)
{
    gint i;
    gchar temp[128];
    
    if (stream->managed) {
        gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets->bitrate_button), TRUE);
        widgets->managed = 1;
    }
    else {
        gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets->quality_button), TRUE);
        widgets->managed = 0;
    }
    
    gtk_adjustment_set_value(GTK_ADJUSTMENT (widgets->quality_adjustment), stream->quality);
    sprintf(temp, "%d", stream->bitrate_min);
    gtk_entry_set_text(GTK_ENTRY (widgets->bitrate_min_entry), temp);
    sprintf(temp, "%d", stream->bitrate);
    gtk_entry_set_text(GTK_ENTRY (widgets->bitrate_nom_entry), temp);
    sprintf(temp, "%d", stream->bitrate_max);
    gtk_entry_set_text(GTK_ENTRY (widgets->bitrate_max_entry), temp);
    
    for (i=0; i<SAMPLERATES; i++) {
        if (oc_samplerates[i] == stream->samplerate) {
            gtk_option_menu_set_history (GTK_OPTION_MENU (widgets->samplerate_menu), i);
            widgets->samplerate = oc_samplerates[i];
        }
    }
    gtk_option_menu_set_history (GTK_OPTION_MENU (widgets->channels_menu),stream->channels-1);
    widgets->channels = stream->channels;
    
    gtk_entry_set_text(GTK_ENTRY (widgets->server_entry), stream->hostname);
    sprintf(temp, "%d", stream->port);
    gtk_entry_set_text(GTK_ENTRY (widgets->port_entry), temp);
    gtk_entry_set_text(GTK_ENTRY (widgets->user_entry), stream->user);
    gtk_entry_set_text(GTK_ENTRY (widgets->password_entry), stream->password);
    gtk_entry_set_text(GTK_ENTRY (widgets->mount_entry), stream->mount);

    gtk_entry_set_text(GTK_ENTRY (widgets->stream_name_entry), stream->name);
    gtk_entry_set_text(GTK_ENTRY (widgets->stream_genre_entry), stream->genre);
    gtk_entry_set_text(GTK_ENTRY (widgets->stream_description_entry), stream->description);
}

void get_values(oc_stream_info *stream)
{
    gchar *temp;
    
    if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widgets->bitrate_button)))
        stream->managed = TRUE;
    else
        stream->managed = FALSE;
    
    stream->quality = GTK_ADJUSTMENT (widgets->quality_adjustment)->value;
    temp = gtk_entry_get_text(GTK_ENTRY (widgets->bitrate_min_entry));
    stream->bitrate_min = atoi(temp);
    temp = gtk_entry_get_text(GTK_ENTRY (widgets->bitrate_nom_entry));
    stream->bitrate = atoi(temp);
    temp = gtk_entry_get_text(GTK_ENTRY (widgets->bitrate_max_entry));
    stream->bitrate_max = atoi(temp);
    
    stream->samplerate = widgets->samplerate;
    stream->channels = widgets->channels;
    
    if (stream->hostname) g_free(stream->hostname);
    if (stream->user) g_free(stream->user);
    if (stream->password) g_free(stream->password);
    if (stream->mount) g_free(stream->mount);
    
    stream->hostname = g_strdup(gtk_entry_get_text(GTK_ENTRY (widgets->server_entry)));
    temp = gtk_entry_get_text(GTK_ENTRY (widgets->port_entry));
    stream->port = atoi(temp);
    stream->user = g_strdup(gtk_entry_get_text(GTK_ENTRY (widgets->user_entry)));
    stream->password = g_strdup(gtk_entry_get_text(GTK_ENTRY (widgets->password_entry)));
    stream->mount = g_strdup(gtk_entry_get_text(GTK_ENTRY (widgets->mount_entry)));
    
    if (stream->name) g_free(stream->name);
    if (stream->genre) g_free(stream->genre);
    if (stream->description) g_free(stream->description);
    
    stream->name = g_strdup(gtk_entry_get_text(GTK_ENTRY (widgets->stream_name_entry)));
    stream->genre = g_strdup(gtk_entry_get_text(GTK_ENTRY (widgets->stream_genre_entry)));
    stream->description = g_strdup(gtk_entry_get_text(GTK_ENTRY (widgets->stream_description_entry)));
}

GtkWidget *create_encoding_frame(void)
{
    GtkWidget *encoding_frame;
    GtkWidget *encoding_hbox;
    GtkWidget *quality_vbox;
    GtkWidget *bitrate_vbox;
    GtkWidget *scale;
    GtkWidget *bitrate_min_label, *bitrate_nom_label, *bitrate_max_label;
    GtkWidget *bitrate_table;
    GSList *group;
    
    encoding_frame = gtk_frame_new ("Encoding Settings");
    
    encoding_hbox = gtk_hbox_new (FALSE, 0);
    gtk_container_add (GTK_CONTAINER (encoding_frame), encoding_hbox);
    gtk_widget_show (encoding_hbox);
    
    quality_vbox = gtk_vbox_new (FALSE, 0);
    gtk_box_pack_start(GTK_BOX(encoding_hbox), quality_vbox, TRUE, TRUE, 10);
    gtk_widget_show (quality_vbox);
    
    widgets->quality_button = gtk_radio_button_new_with_label (NULL, "Quality:");
    group = gtk_radio_button_group (GTK_RADIO_BUTTON (widgets->quality_button));
    gtk_signal_connect(GTK_OBJECT(widgets->quality_button), "toggled",
              GTK_SIGNAL_FUNC(set_managed_cb), GINT_TO_POINTER (1));
    gtk_box_pack_start(GTK_BOX(quality_vbox), widgets->quality_button, FALSE, FALSE, 10);
    gtk_widget_show (widgets->quality_button);
    
    widgets->quality_adjustment = gtk_adjustment_new(3.0, 0.0, 10.0 + 1.0, 0.1, 1.0, 1.0);
    scale = gtk_hscale_new(GTK_ADJUSTMENT(widgets->quality_adjustment));
    gtk_box_pack_start(GTK_BOX(quality_vbox), scale, FALSE, FALSE, 10);
    gtk_widget_show(scale);
    
    bitrate_vbox = gtk_vbox_new (FALSE, 0);
    gtk_box_pack_start(GTK_BOX(encoding_hbox), bitrate_vbox, FALSE, FALSE, 10);
    gtk_widget_show (bitrate_vbox);
    
    widgets->bitrate_button = gtk_radio_button_new_with_label (group, "Bitrate:");
    gtk_signal_connect(GTK_OBJECT(widgets->bitrate_button), "toggled",
              GTK_SIGNAL_FUNC(set_managed_cb), GINT_TO_POINTER (0));
    gtk_box_pack_start(GTK_BOX(bitrate_vbox), widgets->bitrate_button, FALSE, FALSE, 10);
    gtk_widget_show (widgets->bitrate_button);
    
    bitrate_table = gtk_table_new (2, 3, FALSE);
    gtk_container_set_border_width (GTK_CONTAINER (bitrate_table), 5);
    gtk_box_pack_start(GTK_BOX(bitrate_vbox), bitrate_table, FALSE, FALSE, 10);
    gtk_widget_show (bitrate_table);
    
    bitrate_min_label = gtk_label_new ("Minimum:");
    gtk_misc_set_alignment(GTK_MISC(bitrate_min_label), 1, 0.5);
    gtk_table_attach (GTK_TABLE (bitrate_table), bitrate_min_label, 0, 1, 0, 1,
              (GtkAttachOptions) (GTK_FILL),
              (GtkAttachOptions) (0), 5, 5);
    gtk_widget_show (bitrate_min_label);
    
    widgets->bitrate_min_entry = gtk_entry_new ();
    gtk_table_attach (GTK_TABLE (bitrate_table), widgets->bitrate_min_entry, 1, 2, 0, 1,
              (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
              (GtkAttachOptions) (0), 0, 5);
    gtk_widget_show (widgets->bitrate_min_entry);
    
    bitrate_nom_label = gtk_label_new ("Nominal:");
    gtk_misc_set_alignment(GTK_MISC(bitrate_nom_label), 1, 0.5);
    gtk_table_attach (GTK_TABLE (bitrate_table), bitrate_nom_label, 0, 1, 1, 2,
              (GtkAttachOptions) (GTK_FILL),
              (GtkAttachOptions) (0), 5, 5);
    gtk_widget_show (bitrate_nom_label);
    
    widgets->bitrate_nom_entry = gtk_entry_new ();
    gtk_table_attach (GTK_TABLE (bitrate_table), widgets->bitrate_nom_entry, 1, 2, 1, 2,
              (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
              (GtkAttachOptions) (0), 0, 5);
    gtk_widget_show (widgets->bitrate_nom_entry);
    
    bitrate_max_label = gtk_label_new ("Maximum:");
    gtk_misc_set_alignment(GTK_MISC(bitrate_max_label), 1, 0.5);
    gtk_table_attach (GTK_TABLE (bitrate_table), bitrate_max_label, 0, 1, 2, 3,
              (GtkAttachOptions) (GTK_FILL),
              (GtkAttachOptions) (0), 5, 5);
    gtk_widget_show (bitrate_max_label);
    
    widgets->bitrate_max_entry = gtk_entry_new ();
    gtk_table_attach (GTK_TABLE (bitrate_table), widgets->bitrate_max_entry, 1, 2, 2, 3,
              (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
              (GtkAttachOptions) (0), 0, 5);
    gtk_widget_show (widgets->bitrate_max_entry);
    
    return encoding_frame;
}

GtkWidget *create_format_frame(void)
{
    GtkWidget *format_frame;
    GtkWidget *format_hbox;
    GtkWidget *samplerate_label;
    GtkWidget *samplerate_menu_menu;
    GtkWidget *channels_label;
    GtkWidget *channels_menu_menu;
    GtkWidget *menuitem;
    
    gint i;
    gchar temp[128];
    
    format_frame = gtk_frame_new ("Audio Format");
    
    format_hbox = gtk_hbox_new (FALSE, 0);
    gtk_container_add (GTK_CONTAINER (format_frame), format_hbox);
    gtk_widget_show (format_hbox);
    
    samplerate_label = gtk_label_new ("Sample Rate (Hz):");
    gtk_box_pack_start(GTK_BOX(format_hbox), samplerate_label, FALSE, FALSE, 10);
    gtk_widget_show (samplerate_label);
    
    widgets->samplerate_menu = gtk_option_menu_new ();    
    gtk_box_pack_start(GTK_BOX(format_hbox), widgets->samplerate_menu, TRUE, TRUE, 10);
    gtk_widget_show (widgets->samplerate_menu);
	
    samplerate_menu_menu = gtk_menu_new ();
    for(i=0; i<SAMPLERATES; i++){
        sprintf(temp, "%d", oc_samplerates[i]);
        menuitem = gtk_menu_item_new_with_label(temp);
        gtk_widget_show(menuitem);
        gtk_menu_append (GTK_MENU (samplerate_menu_menu), menuitem);
        
        gtk_signal_connect(GTK_OBJECT(menuitem), "activate",
                   GTK_SIGNAL_FUNC(set_samplerate_cb), GINT_TO_POINTER(oc_samplerates[i]));
    }    
    gtk_option_menu_set_menu (GTK_OPTION_MENU (widgets->samplerate_menu), samplerate_menu_menu);
    
    channels_label = gtk_label_new ("Number Of Channels:");
    gtk_box_pack_start(GTK_BOX(format_hbox), channels_label, FALSE, FALSE, 10);
    gtk_widget_show (channels_label);
    
    widgets->channels_menu = gtk_option_menu_new ();
    gtk_box_pack_start(GTK_BOX(format_hbox), widgets->channels_menu, TRUE, TRUE, 10);
    gtk_widget_show (widgets->channels_menu);
    
    channels_menu_menu = gtk_menu_new ();
    menuitem = gtk_menu_item_new_with_label ("Mono");
    gtk_signal_connect(GTK_OBJECT(menuitem), "activate",
                   GTK_SIGNAL_FUNC(set_channels_cb), GINT_TO_POINTER(1));
    gtk_widget_show (menuitem);
    gtk_menu_append (GTK_MENU (channels_menu_menu), menuitem);
    menuitem = gtk_menu_item_new_with_label ("Stereo");
    gtk_signal_connect(GTK_OBJECT(menuitem), "activate",
                   GTK_SIGNAL_FUNC(set_channels_cb), GINT_TO_POINTER(2));
    gtk_widget_show (menuitem);
    gtk_menu_append (GTK_MENU (channels_menu_menu), menuitem);
    gtk_option_menu_set_menu (GTK_OPTION_MENU (widgets->channels_menu), channels_menu_menu);
    
    return format_frame;
}

GtkWidget *create_server_frame(void)
{
    GtkWidget *server_frame;
    GtkWidget *server_table;
    GtkWidget *server_label, *port_label, *user_label, *password_label, *mount_label;
    
    server_frame = gtk_frame_new ("Server Settings");
    
    server_table = gtk_table_new (4, 3, FALSE);
    gtk_container_set_border_width (GTK_CONTAINER (server_table), 5);
    gtk_container_add (GTK_CONTAINER (server_frame), server_table);
    gtk_widget_show (server_table);
    
    server_label = gtk_label_new ("Address:");
    gtk_misc_set_alignment(GTK_MISC(server_label), 1, 0.5);
    gtk_table_attach (GTK_TABLE (server_table), server_label, 0, 1, 0, 1,
              (GtkAttachOptions) (GTK_FILL),
              (GtkAttachOptions) (0), 5, 5);
    gtk_widget_show (server_label);
    
    widgets->server_entry = gtk_entry_new ();
    gtk_table_attach (GTK_TABLE (server_table), widgets->server_entry, 1, 2, 0, 1,
              (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
              (GtkAttachOptions) (0), 0, 5);
    gtk_widget_show (widgets->server_entry);
    
    port_label = gtk_label_new ("Port:");
    gtk_misc_set_alignment(GTK_MISC(port_label), 1, 0.5);
    gtk_table_attach (GTK_TABLE (server_table), port_label, 2, 3, 0, 1,
              (GtkAttachOptions) (GTK_FILL),
              (GtkAttachOptions) (0), 5, 5);
    gtk_widget_show (port_label);
    
    widgets->port_entry = gtk_entry_new_with_max_length (5);
    gtk_table_attach (GTK_TABLE (server_table), widgets->port_entry, 3, 4, 0, 1,
              (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
              (GtkAttachOptions) (0), 0, 5);
    gtk_widget_show (widgets->port_entry);
    
    user_label = gtk_label_new ("Username:");
    gtk_misc_set_alignment(GTK_MISC(user_label), 1, 0.5);
    gtk_table_attach (GTK_TABLE (server_table), user_label, 0, 1, 1, 2,
              (GtkAttachOptions) (GTK_FILL),
              (GtkAttachOptions) (0), 5, 5);
    gtk_widget_show (user_label);
    
    widgets->user_entry = gtk_entry_new ();
    gtk_table_attach (GTK_TABLE (server_table), widgets->user_entry, 1, 2, 1, 2,
              (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
              (GtkAttachOptions) (0), 0, 5);
    gtk_widget_show (widgets->user_entry);
    
    password_label = gtk_label_new ("Password:");
    gtk_misc_set_alignment(GTK_MISC(password_label), 1, 0.5);
    gtk_table_attach (GTK_TABLE (server_table), password_label, 2, 3, 1, 2,
              (GtkAttachOptions) (GTK_FILL),
              (GtkAttachOptions) (0), 5, 5);
    gtk_widget_show (password_label);
    
    widgets->password_entry = gtk_entry_new ();
    gtk_entry_set_visibility(GTK_ENTRY(widgets->password_entry), FALSE);
    gtk_table_attach (GTK_TABLE (server_table), widgets->password_entry, 3, 4, 1, 2,
              (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
              (GtkAttachOptions) (0), 0, 5);
    gtk_widget_show (widgets->password_entry);
    
    mount_label = gtk_label_new ("Mount:");
    gtk_misc_set_alignment(GTK_MISC(mount_label), 1, 0.5);
    gtk_table_attach (GTK_TABLE (server_table), mount_label, 0, 1, 2, 3,
              (GtkAttachOptions) (GTK_FILL),
              (GtkAttachOptions) (0), 5, 5);
    gtk_widget_show (mount_label);
    
    widgets->mount_entry = gtk_entry_new ();
    gtk_table_attach (GTK_TABLE (server_table), widgets->mount_entry, 1, 2, 2, 3,
              (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
              (GtkAttachOptions) (0), 0, 5);
    gtk_widget_show (widgets->mount_entry);
    
    return server_frame;
}

GtkWidget *create_info_frame(void) 
{
    GtkWidget *info_frame;
    GtkWidget *info_table;
    GtkWidget *name_label, *genre_label, *description_label;
    
    info_frame = gtk_frame_new ("Stream Info");
    
    info_table = gtk_table_new (2, 3, FALSE);
    gtk_container_set_border_width (GTK_CONTAINER (info_table), 5);
    gtk_container_add (GTK_CONTAINER (info_frame), info_table);
    gtk_widget_show (info_table);
    
    name_label = gtk_label_new ("Name:");
    gtk_misc_set_alignment(GTK_MISC(name_label), 1, 0.5);
    gtk_table_attach (GTK_TABLE (info_table), name_label, 0, 1, 0, 1,
              (GtkAttachOptions) (GTK_FILL),
              (GtkAttachOptions) (0), 5, 5);
    gtk_widget_show (name_label);
    
    widgets->stream_name_entry = gtk_entry_new ();
    gtk_table_attach (GTK_TABLE (info_table), widgets->stream_name_entry, 1, 2, 0, 1,
              (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
              (GtkAttachOptions) (0), 0, 5);
    gtk_widget_show (widgets->stream_name_entry);
    
    genre_label = gtk_label_new ("Genre:");
    gtk_misc_set_alignment(GTK_MISC(genre_label), 1, 0.5);
    gtk_table_attach (GTK_TABLE (info_table), genre_label, 0, 1, 1, 2,
              (GtkAttachOptions) (GTK_FILL),
              (GtkAttachOptions) (0), 5, 5);
    gtk_widget_show (genre_label);
    
    widgets->stream_genre_entry = gtk_entry_new ();
    gtk_table_attach (GTK_TABLE (info_table), widgets->stream_genre_entry, 1, 2, 1, 2,
              (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
              (GtkAttachOptions) (0), 0, 5);
    gtk_widget_show (widgets->stream_genre_entry);
    
    description_label = gtk_label_new ("Description:");
    gtk_misc_set_alignment(GTK_MISC(description_label), 1, 0.5);
    gtk_table_attach (GTK_TABLE (info_table), description_label, 0, 1, 2, 3,
              (GtkAttachOptions) (GTK_FILL),
              (GtkAttachOptions) (0), 5, 5);
    gtk_widget_show (description_label);
    
    widgets->stream_description_entry = gtk_entry_new ();
    gtk_table_attach (GTK_TABLE (info_table), widgets->stream_description_entry, 1, 2, 2, 3,
              (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
              (GtkAttachOptions) (0), 0, 5);
    gtk_widget_show (widgets->stream_description_entry);
    
    return info_frame;
}

GtkWidget *create_buttons(void)
{
    GtkWidget *buttons_vbox;
    GtkWidget *control_hbox, *close_hbox;
    
    buttons_vbox = gtk_vbox_new (TRUE, 0);
    control_hbox = gtk_hbox_new (FALSE, 0);
    
    if (widgets->state->running)
        widgets->connect_button = gtk_button_new_with_label ("Disconnect");
    else widgets->connect_button = gtk_button_new_with_label ("Connect");
    gtk_box_pack_start (GTK_BOX (control_hbox), widgets->connect_button, TRUE, TRUE, 0);
    gtk_widget_show (widgets->connect_button);
    gtk_signal_connect( GTK_OBJECT(widgets->connect_button), "clicked", GTK_SIGNAL_FUNC(connect_button_cb), NULL);
    
    widgets->apply_button = gtk_button_new_with_label ("Apply settings");
    gtk_box_pack_start (GTK_BOX (control_hbox), widgets->apply_button, TRUE, TRUE, 30);
    gtk_widget_show (widgets->apply_button);
    gtk_signal_connect( GTK_OBJECT(widgets->apply_button), "clicked", GTK_SIGNAL_FUNC(apply_button_cb), NULL);
    
    widgets->mute_button = gtk_toggle_button_new_with_label ("Mute");
    if (widgets->state->muted) gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON (widgets->mute_button), TRUE);
    gtk_box_pack_start (GTK_BOX (control_hbox), widgets->mute_button, TRUE, TRUE, 0);
    gtk_widget_show (widgets->mute_button);
    gtk_signal_connect( GTK_OBJECT(widgets->mute_button), "clicked", GTK_SIGNAL_FUNC(mute_button_cb), NULL);
    
    gtk_box_pack_start (GTK_BOX (buttons_vbox), control_hbox, FALSE, FALSE, 0);
    gtk_widget_show (control_hbox);
    
    close_hbox = gtk_hbutton_box_new ();
    gtk_button_box_set_layout(GTK_BUTTON_BOX(close_hbox), GTK_BUTTONBOX_END);
    gtk_button_box_set_spacing(GTK_BUTTON_BOX(close_hbox), 10);
    
    widgets->ok_button = gtk_button_new_with_label ("OK");
    gtk_box_pack_start (GTK_BOX (close_hbox), widgets->ok_button, FALSE, FALSE, 0);
    gtk_widget_show (widgets->ok_button);
    gtk_signal_connect( GTK_OBJECT(widgets->ok_button), "clicked", GTK_SIGNAL_FUNC(ok_button_cb), NULL);
    
    widgets->cancel_button = gtk_button_new_with_label ("Cancel");
    gtk_box_pack_start (GTK_BOX (close_hbox), widgets->cancel_button, FALSE, FALSE, 0);
    gtk_widget_show (widgets->cancel_button);

    gtk_box_pack_start (GTK_BOX (buttons_vbox), close_hbox, FALSE, FALSE, 10);
    gtk_widget_show (close_hbox);
    
    return buttons_vbox;
}
