/* oggcaster.c
 * - basic plugin functionality
 *
 * Copyright (c) 2003 Juho Vh-Herttua <juhovh@iki.fi>
 *
 * This program is distributed under the terms of the GNU General
 * Public License, version 2. You may use, modify, and redistribute
 * it under the terms of this license. A copy should be included
 * with this source.
 */

#ifdef HAVE_CONFIG_H
 #include <config.h>
#endif

#include <stdlib.h>
#include <glib.h>
#include <ogg/ogg.h>
#include <vorbis/codec.h>
#include <xmms/plugin.h>
#include <xmms/xmmsctrl.h>
#include <xmms/util.h>
#include "charconv/utf8.h"
#include "oggcaster.h"
#include "configure.h"
#include "resample.h"
#include "encode.h"
#include "stream.h"
#include "audio.h"
#include "gui.h"

static void oggcaster_init(void);
static void oggcaster_cleanup(void);
static void oggcaster_configure(void);
static void oggcaster_about(void);
static int mod_samples(gpointer *data, gint length, AFormat fmt, gint srate, gint nch);
static void query_format(AFormat *fmt, gint *rate, gint *nch);
static gint timeout_func(gpointer data);

static oc_state_info *state;

EffectPlugin oggcaster_ep = {
    0, // Handle, filled in by xmms
    0, // Filename, filled in by xmms
    "OggCaster Plugin " VERSION, /* Description */
    oggcaster_init,
    oggcaster_cleanup,
    oggcaster_about,
    oggcaster_configure,
    mod_samples,
    query_format
};

EffectPlugin *get_eplugin_info(void)
{
	return &oggcaster_ep;
}

static void oggcaster_init(void) 
{
    state = g_malloc0(sizeof(oc_state_info));
    state->stream = g_malloc0(sizeof(oc_stream_info));
    state->proc = process_initialise();
    
    /* Load previously saved settings */
    state->saved_stream = config_load();
    
    /* FIXME: timeout should be made optional and charset configurable */
    state->timeout_tag = gtk_timeout_add(100, timeout_func, NULL);
    convert_set_charset("ISO-8859-1");

    pthread_mutex_init(&state->runlock, NULL);
}

static void oggcaster_cleanup(void)
{
    /* Save current settings */
    config_save(state->saved_stream);
    
    /* Wait for the encoding thread to stop */
    pthread_mutex_lock(&state->runlock);
    if (state->running) {
        state->running = FALSE;
        stream_queue(state, NULL, 0);
        pthread_mutex_unlock(&state->runlock);
        pthread_join(state->thread, NULL);
    }
    else pthread_mutex_unlock(&state->runlock);
    pthread_mutex_destroy(&state->runlock);
    
    /* Clear timeout */
    if (state->timeout_tag)
        gtk_timeout_remove(state->timeout_tag);
    
    /* Clear audio processing */
    process_clear(state->proc);
    
    /* Free all used structs */
    g_free(state->saved_stream);
    g_free(state->stream);
    g_free(state);
}

static void oggcaster_configure(void)
{
    /* We do this in gui.c */
    configure(state);
}

static void oggcaster_about(void)
{
	static GtkWidget *dialog;
    
	if (dialog != NULL)
		return;
	
	dialog = xmms_show_message(
		("About OggCaster Plugin"),
		("XMMS OggCaster Plugin\n\n"
		 "An icecast2 client plugin by "
		 "Juho Vaha-Herttua 2003"),
		("Ok"), FALSE, NULL, NULL);

	gtk_signal_connect(GTK_OBJECT(dialog), "destroy",
			   GTK_SIGNAL_FUNC(gtk_widget_destroyed),
			   &dialog);
}

static int mod_samples(gpointer *d, gint length, AFormat fmt, gint srate, gint channels)
{
    gint i, lowbits=0, usigned=0, be=0;
    gfloat **buffer=NULL;
    gint bufsize;
    gint8 *data = (gint8 *) *d;
    
    /* If we are not streaming just return (and mute) */
    pthread_mutex_lock(&state->runlock);
    if (!state->running) {
        if (state->muted)
            for (i=0; i<length; i++)
                data[i] = 0;
        pthread_mutex_unlock(&state->runlock);
        return length;
    }
    
    /* Convert current fmt to correct input settings */
    if (fmt == FMT_U8 || fmt == FMT_S8)
        lowbits = 1;
    if (fmt == FMT_U8 || fmt == FMT_U16_NE || fmt == FMT_U16_LE || fmt == FMT_U16_BE)
        usigned = 1;
    if (fmt == FMT_U16_BE || fmt == FMT_S16_BE || (fmt == FMT_U16_NE && G_BYTE_ORDER == G_BIG_ENDIAN))
        be = 1;
    
    /* If we have different input settings than before, initialise new processor */
    if (fmt != state->old_format || srate != state->old_srate ||
            channels != state->old_channels || state->new_stream) {
        
        bufsize = process_finish(state->proc, &buffer);
        
        /* Don't send data if we have a new connection, just flush */
        if (bufsize != 0 && !state->new_connection) {
            stream_queue(state, buffer, bufsize);
            state->new_connection = FALSE;
        }
        
        /* Update processing settings and state settings */
        process_update(state->proc, lowbits, usigned, be, channels, state->stream->channels, srate, state->stream->samplerate);
        state->old_format = fmt;
        state->old_srate = srate;
        state->old_channels = channels;
        
        /* If this was because of a change in output, initialise new encoder */
        if (state->new_stream) {
            state->new_stream = FALSE;
            state->new_encoder = TRUE;
        }
    }
    
    /* Process and send input audio */
    bufsize = process_audio(state->proc, data, length, &buffer);
    stream_queue(state, buffer, bufsize);
    
    /* If we are muted then just return nothing in data buffer */
    if (state->muted)
        for (i=0; i<length; i++)
            data[i] = 0;
    
    /* Unlock and return */
    pthread_mutex_unlock(&state->runlock);
    return length;    
}

static void query_format(AFormat *fmt, gint *rate, gint *nch)
{
    pthread_mutex_lock(&state->runlock);
    if (state->running) {
        /* We are using only signed int, not unsigned... this saves us one buffer */
        if (*fmt == FMT_U8) *fmt = FMT_S8;
        else if (*fmt != FMT_S8) *fmt = FMT_S16_LE;
    }
    pthread_mutex_unlock(&state->runlock);
}

static gint timeout_func(gpointer data)
{
    gint pos, length;
    gboolean playing;
    gchar *title;
    gchar *utftitle;
    vorbis_comment *comment;
    
    /* If we are not running we don't need to do anything */
    if (!state->running) return TRUE;
    
    gdk_threads_enter();
    
    playing = xmms_remote_is_playing(0);
    pos = xmms_remote_get_playlist_pos(0);
    length = xmms_remote_get_playlist_time(0, pos);
    
    /* If settings have changed and we don't have a comment, make one */
    if ((pos != state->previous_song || length != state->previous_length) &&
            state->vc == NULL) {
        state->previous_song = pos;
        state->previous_length = length;
        title = xmms_remote_get_playlist_title(0, pos);
        if (title) {
            /* Initialise a new vorbis comment and add needed tags encoded in UTF-8 */
            pthread_mutex_lock(&state->runlock);
            utf8_encode(title, &utftitle);
            comment = g_malloc0(sizeof(vorbis_comment));
            vorbis_comment_init(comment);
            vorbis_comment_add_tag(comment, "title", utftitle);
            g_free(utftitle);
            utf8_encode(state->stream->name, &utftitle);
            vorbis_comment_add_tag(comment, "artist", utftitle);
            state->vc = comment;
            g_free(utftitle);
            g_free(title);
            
            /* We also want a new encoder here */
            state->new_encoder = TRUE;
            pthread_mutex_unlock(&state->runlock);
        }
    }
    
    gdk_threads_leave();
    
    return TRUE;
}
