/* stream.c
 * - audio encoding and broadcasting
 *
 * Copyright (c) 2003 Juho Vh-Herttua <juhovh@iki.fi>
 *
 * This program is distributed under the terms of the GNU General
 * Public License, version 2. You may use, modify, and redistribute
 * it under the terms of this license. A copy should be included
 * with this source.
 */
 
#ifdef HAVE_CONFIG_H
 #include <config.h>
#endif

#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <glib.h>
#include <pthread.h>
#include <shout/shout.h>
#include <ogg/ogg.h>
#include <vorbis/codec.h>
#include <assert.h>
#include "oggcaster.h"
#include "encode.h"
#include "stream.h"
#include "gui.h"

#define MAX_BUFSIZE 2048*1024

void *encodingThread(void *threadarg)
{
    gint i, ret, shoutret;
    gboolean exit=FALSE;
    ogg_page og;
    shout_t *shout = NULL;
    encoder_state *tempenc, *enc = NULL;
    ref_buffer *buffer;
    queue_item *old;
    vorbis_comment *comment;
    oc_state_info *state = (oc_state_info *) threadarg;
    oc_stream_info *stream = state->stream;
    buffer_queue *queue = state->queue;
    
    comment = g_malloc0(sizeof(vorbis_comment));
    vorbis_comment_init(comment);
    shout = shout_new();
    
    /* just check if the settings are legal and allowed */
    tempenc = encode_initialise(stream->channels, stream->samplerate,
                                stream->managed, stream->bitrate_min,
                                stream->bitrate, stream->bitrate_max,
                                stream->quality, stream->serial++, comment);
    encode_clear(tempenc);
    
    if (tempenc == NULL)
        ret = 4;
    else if (shout_initialise(shout, stream) == 0)
        ret = 3;
    else if (!(shout_open(shout)) == SHOUTERR_SUCCESS)
        ret = 2;
    else
        ret = 1;
    
    connect_button_activate(ret);
    
    while (!exit) {
        pthread_mutex_lock(&queue->lock);
        if (queue->head == NULL)
            pthread_cond_wait(&state->cond, &queue->lock);
        
        buffer = queue->head->buf;
        old = queue->head;
        
        queue->head = queue->head->next;
        if(!queue->head)
            queue->tail = NULL;
        
        g_free(old);
        queue->length -= buffer->len * buffer->channels * sizeof(gfloat);
        pthread_mutex_unlock(&queue->lock);
        
        exit = buffer->kill;
        
        if (buffer->new_encoder || enc == NULL) {
            if (buffer->vc) {
                vorbis_comment_clear(comment);
                g_free(comment);
                comment = buffer->vc;
            }
            
            tempenc = encode_initialise(stream->channels, stream->samplerate,
                                    stream->managed, stream->bitrate_min,
                                    stream->bitrate, stream->bitrate_max,
                                    stream->quality, stream->serial++, comment);
            if (tempenc) {
                if (enc) {
                    encode_finish(enc);
                    while(encode_flush(enc, &og) != 0) {
                        shout_send_raw(shout, og.header, og.header_len);
                        shout_send_raw(shout, og.body, og.body_len);
                    }
                    encode_clear(enc);
                }
                enc = tempenc;
            }
            else {
                /* Error in initialising encoder, we disconnect and return */
                if (enc)
                    encode_finish(enc);
                pthread_mutex_lock(&state->runlock);
                state->running = FALSE;
                stream_flush(queue);
                pthread_mutex_unlock(&state->runlock);
                
                shout_close(shout);
                shout_free(shout);
                ret = 4;
                connect_button_activate(ret);
                exit = TRUE;
            }
        }
        
        if (!exit) {
            encode_data_float(enc, buffer->buf, buffer->len);
            
            while(encode_dataout(enc, &og) > 0) {
                if ((shoutret = shout_send_raw(shout, og.header, og.header_len)) == 0)
                    break;
                if ((shoutret = shout_send_raw(shout, og.body, og.body_len)) == 0);
                    break;
            }
            if (shoutret <= 0) {
                /* Some problems with connection, we just exit now */
                pthread_mutex_lock(&state->runlock);
                state->running = FALSE;
                stream_flush(queue);
                pthread_mutex_unlock(&state->runlock);
                
                shout_close(shout);
                shout_free(shout);
                encode_clear(enc);
                ret = 5;
                connect_button_activate(ret);
                exit = TRUE;
            }
        }
        
        for (i=0; i<buffer->channels; i++)
            g_free(buffer->buf[i]);
        g_free(buffer->buf);
        g_free(buffer);
    }
    
    /* Close and free connection if needed */
    switch(ret) {
        case 1:
            shout_close(shout);
        case 2:
            shout_free(shout);
            break;
    }
    g_free(queue);
    
    pthread_cond_destroy(&state->cond);
    pthread_mutex_destroy(&queue->lock);
    
    /* Return information to gui and exit thread */
    if (ret == 1)
        connect_button_activate(0);
    pthread_exit((void *) 0);
}

void stream_initialise(oc_state_info *state)
{
    pthread_attr_t attr;
    
    state->queue = g_malloc0(sizeof(buffer_queue));
    
    /* Generate new serial number for stream */
    srand(time(NULL));
    state->stream->serial = rand();
    
    /* We have a new connection and a new stream here (new encoder later) */
    state->new_connection = TRUE;
    state->new_stream = TRUE;
    state->new_encoder = FALSE;
	state->previous_song = -1;
	state->previous_length = -1;
    
    /* Initialise thread settings and start thread */
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
    pthread_mutex_init(&state->queue->lock, NULL);
    pthread_cond_init(&state->cond, NULL);
    if (pthread_create(&state->thread, &attr, encodingThread, (void *) state))
        g_print("Error initialising thread\n");
    pthread_attr_destroy(&attr);
}

void stream_queue(oc_state_info *state, gfloat **buffer, gint buflen)
{
    buffer_queue *queue;
    ref_buffer *chunk;
    queue_item *queued;
    gint i;
    
    queue = state->queue;
    
    /* Allocate new buffer chunk */
    chunk = g_malloc0(sizeof(ref_buffer));
    queued = g_malloc(sizeof(queue_item));
    queued->buf = chunk;
    chunk->channels = state->stream->channels;
    chunk->buf = g_malloc0(sizeof(gfloat *)*chunk->channels);
    
    /* If we have any input, duplicate it to the buffer chunk */
    if (buflen) {
        for (i=0; i<chunk->channels; i++)
            chunk->buf[i] = g_memdup(buffer[i], sizeof(gfloat)*buflen);
    }
    chunk->len = buflen;
    
    /* If we want a new encoder, inform the thread */
    if (state->new_encoder) {
        if (state->vc) {
            chunk->vc = state->vc;
            state->vc = NULL;
        }
        chunk->new_encoder = TRUE;
        state->new_encoder = FALSE;
    }
    
    /* If we are not running any more, inform the thread (called from gui.c) */
    if (!state->running)
        chunk->kill = TRUE;
    
    /* Put current chunk in queue */
    pthread_mutex_lock(&queue->lock);
    pthread_cond_signal(&state->cond);
    if(queue->head == NULL) {
        queue->head = queue->tail = queued;
        queue->head->next = queue->tail->next = NULL;
    }
    else {
        queue->tail->next = queued;
        queued->next = NULL;
        queue->tail = queued;
    }
    queue->length += chunk->len * chunk->channels * sizeof(gfloat);

    /* While we have exceeded the MAX_BUFSIZE, flush from the beginning */
    while (queue->length > MAX_BUFSIZE) {
        chunk = queue->head->buf;
        queued = queue->head;
        queue->head = queue->head->next;
        if(!queue->head)
            queue->tail = NULL;
        queue->length -= chunk->len * chunk->channels * sizeof(gfloat);
        
        g_free(queued);
        for (i=0; i<chunk->channels; i++)
            g_free(chunk->buf[i]);
        g_free(chunk->buf);
        g_free(chunk);
    }
    pthread_mutex_unlock(&queue->lock);
}

void stream_flush(buffer_queue *queue)
{
    ref_buffer *buffer;
    queue_item *old;
    gint i;
    
    /* While we have something in the buffer, just flush it */
    while (queue->head != NULL) {
        buffer = queue->head->buf;
        old = queue->head;
    
        queue->head = queue->head->next;
        if(!queue->head)
            queue->tail = NULL;
    
        g_free(old);
        queue->length -= buffer->len * buffer->channels * sizeof(gfloat);
        for (i=0; i<buffer->channels; i++)
            g_free(buffer->buf[i]);
        g_free(buffer->buf);
        g_free(buffer);
    }
}

/* Initialise the shout_t, no error handling yet */
gint shout_initialise(shout_t *shout, oc_stream_info *stream)
{
    gchar *user;
    
    shout_set_format(shout, SHOUT_FORMAT_VORBIS);
    shout_set_protocol(shout, SHOUT_PROTOCOL_HTTP);
    
    if (!(shout_set_host(shout, stream->hostname)) == SHOUTERR_SUCCESS) {
        //g_print("libshout error: %s\n", shout_get_error(shout));
        return 0;
    }
    if (!(shout_set_port(shout, stream->port)) == SHOUTERR_SUCCESS) {
        return 0;
    }
    
    if (stream->user) user = stream->user;
    else user = "source";
    
    if (!(shout_set_user(shout, user)) == SHOUTERR_SUCCESS) {
        return 0;
    }
    if (!(shout_set_password(shout, stream->password)) == SHOUTERR_SUCCESS) {
        return 0;
    }
    if (!(shout_set_agent(shout, "OggCaster XMMS plugin")) == SHOUTERR_SUCCESS) {
        return 0;
    }
    if (!(shout_set_mount(shout, stream->mount)) == SHOUTERR_SUCCESS) {
        return 0;
    }
    
    // Metadata settings here
    if (stream->name)
        if (!(shout_set_name(shout, stream->name)) == SHOUTERR_SUCCESS) {
            return 0;
        }
    if (stream->genre)
        if (!(shout_set_genre(shout, stream->genre)) == SHOUTERR_SUCCESS) {
            return 0;
        }
    if (stream->description)
        if (!(shout_set_description(shout, stream->description)) == SHOUTERR_SUCCESS) {
            return 0;
        }
    return 1;
}
